import os

import grpc

from .api_key import get_api_key
from .v1.baserun_pb2_grpc import SubmissionServiceStub


def credentials() -> grpc.ChannelCredentials:
    if key_chain := os.environ.get("SSL_KEY_CHAIN"):
        ssl_creds = grpc.ssl_channel_credentials(
            root_certificates=bytes(key_chain, "utf-8")
        )
    else:
        ssl_creds = grpc.ssl_channel_credentials()

    call_credentials = grpc.access_token_call_credentials(get_api_key())
    return grpc.composite_channel_credentials(ssl_creds, call_credentials)


def get_or_create_submission_service() -> SubmissionServiceStub:
    from baserun import Baserun

    if Baserun.submission_service is None:
        grpc_base = os.environ.get("BASERUN_GRPC_URI", "grpc.baserun.ai:50051")
        grpc_channel = grpc.secure_channel(grpc_base, credentials())
        Baserun.submission_service = SubmissionServiceStub(grpc_channel)

    return Baserun.submission_service


def get_or_create_async_submission_service() -> SubmissionServiceStub:
    from baserun import Baserun

    if Baserun.async_submission_service is None:
        grpc_base = os.environ.get("BASERUN_GRPC_URI", "grpc.baserun.ai:50051")
        grpc_channel = grpc.aio.secure_channel(grpc_base, credentials())
        Baserun.async_submission_service = SubmissionServiceStub(grpc_channel)

    return Baserun.async_submission_service



#!/usr/bin/env python
import collections.abc
import logging
from typing import Collection

from opentelemetry.sdk.trace import _Span

from baserun.instrumentation.base_instrumentor import BaseInstrumentor

logger = logging.getLogger(__name__)

_instruments = ("anthropic >= 0.3.0",)
__version__ = "0.1.0"


class AnthropicInstrumentor(BaseInstrumentor):
    """An instrumentor for Anthropic's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    @staticmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        for value in original_generator:
            AnthropicInstrumentor._handle_generator_value(value, span)

            yield value

    @staticmethod
    async def async_generator_wrapper(original_generator: collections.abc.AsyncIterator, span: _Span):
        logged_failure = False
        async for value in original_generator:
            try:
                AnthropicInstrumentor._handle_generator_value(value, span)
            except Exception as e:
                # Avoid spamming failures for each token
                if not logged_failure:
                    logger.info(f"Baserun couldn't handle generator value {value}: {e}")
                    logged_failure = True

            yield value

    @staticmethod
    def _handle_generator_value(value, span: _Span):
        new_content = value.completion

        prefix = f"0"
        content_attribute = f"{prefix}.content"

        if new_content:
            content = span.attributes.get(content_attribute, "")
            span.set_attribute(content_attribute, content + new_content)

        if new_content is None or value.stop_reason and span.is_recording():
            span.end()



import hashlib
import inspect
import logging
import os
import traceback
from typing import Any, TYPE_CHECKING, Union, Set

from baserun import Baserun
from baserun.grpc import (
    get_or_create_submission_service,
    get_or_create_async_submission_service,
)
from baserun.helpers import memoize_for_time
from baserun.v1.baserun_pb2 import (
    Template,
    TemplateVersion,
    SubmitTemplateVersionRequest,
    SubmitTemplateVersionResponse,
    GetTemplatesRequest,
    GetTemplatesResponse,
    TemplateMessage,
)

if TYPE_CHECKING:
    # Just for type annotations for Langchain. Since langchain is an optional dependency we have this garbage
    try:
        from langchain.tools import Tool
    except ImportError:
        Tool = Any

logger = logging.getLogger(__name__)


@memoize_for_time(os.environ.get("BASERUN_CACHE_INTERVAL", 600))
def get_templates(environment: str = None) -> dict[str, Template]:
    if not Baserun.templates:
        Baserun.templates = {}

    try:
        request = GetTemplatesRequest(environment=environment or Baserun.environment)
        response: GetTemplatesResponse = get_or_create_submission_service().GetTemplates(request)
        for template in response.templates:
            Baserun.templates[template.name] = template

    except BaseException as e:
        logger.error(f"Could not fetch templates from Baserun. Using {len(Baserun.templates.keys())} cached templates")
        logger.info(traceback.format_exception(e))

    return Baserun.templates


def get_template(name: str) -> Template:
    templates = get_templates()
    template = templates.get(name)
    if not template:
        logger.info(f"Attempted to get template {name} but no template with that name exists")
        return None

    return template


def get_template_type_enum(template_type: str = None):
    template_type = template_type or "unknown"
    if template_type == Template.TEMPLATE_TYPE_JINJA2 or template_type.lower().startswith("jinja"):
        template_type_enum = Template.TEMPLATE_TYPE_JINJA2
    else:
        template_type_enum = Template.TEMPLATE_TYPE_FORMATTED_STRING

    return template_type_enum


def apply_template(
    template_name: str,
    template_messages: list[dict[str, Union[str, dict[str, Any]]]],
    parameters: dict[str, Any],
    template_type_enum,
) -> str:
    formatted_messages = []
    for message in template_messages:
        template_string = message.get("content")
        if template_string:
            if template_type_enum == Template.TEMPLATE_TYPE_JINJA2:
                try:
                    # noinspection PyUnresolvedReferences
                    from jinja2 import Template as JinjaTemplate

                    template = JinjaTemplate(template_string)
                    return template.render(parameters)
                except ImportError:
                    logger.warning("Cannot render Jinja2 template as jinja2 package is not installed")
                    # TODO: Is this OK? should we raise? or return blank string?
                    return template_string

            formatted_content = template_string.format(**parameters)
            formatted_messages.append({**message, "content": formatted_content})
        else:
            formatted_messages.append(message)

    formatted_template_list: Set[tuple] = Baserun.formatted_templates.get(
        template_name,
    )
    set_value = tuple([message.get("content") for message in formatted_messages])

    if not formatted_template_list:
        formatted_template_list = {set_value}
    else:
        formatted_template_list.add(set_value)

    Baserun.formatted_templates[template_name] = formatted_template_list
    return formatted_messages


def create_langchain_template(
    template_string: str,
    parameters: dict[str, Any] = None,
    template_name: str = None,
    template_tag: str = None,
    template_type: str = None,
    tools: list[Union["Tool", Any]] = None,
):
    from langchain.prompts import PromptTemplate

    parameters = parameters or {}
    input_variables = list(parameters.keys())
    template_type = template_type or "Formatted String"
    tools = tools or []

    if not template_name:
        caller = inspect.stack()[1].function
        template_name = f"{caller}_template"

    if template_type == Template.TEMPLATE_TYPE_JINJA2 or template_type.lower().startswith("jinja"):
        langchain_template = PromptTemplate(
            template=template_string,
            input_variables=input_variables,
            template_format="jinja2",
            tools=tools,
        )

    else:
        langchain_template = PromptTemplate(template=template_string, input_variables=input_variables, tools=tools)

    template_type_enum = get_template_type_enum(template_type)
    template_version = register_template(
        template_string=template_string,
        template_name=template_name,
        template_type=template_type_enum,
        template_tag=template_tag,
    )

    return langchain_template


def format_prompt(
    template_name: str,
    parameters: dict[str, Any],
    template_messages: list[dict[str, Union[str, dict[str, Any]]]] = None,
    template_type: str = None,
):
    template_type_enum = get_template_type_enum(template_type)
    if not template_messages:
        template = get_template(template_name)
        template_messages = [
            {"role": message.role, "content": message.message} for message in template.active_version.template_messages
        ]

    return apply_template(template_name, template_messages, parameters, template_type_enum)


def construct_template_version(
    template_messages: list[dict[str, Union[str, dict[str, Any]]]],
    template_name: str = None,
    template_tag: str = None,
    template_type=Template.TEMPLATE_TYPE_FORMATTED_STRING,
) -> TemplateVersion:
    # Automatically generate a name based on the template's contents
    if not template_name:
        template_name = hashlib.sha256(
            "".join([message.get("content") for message in template_messages]).encode()
        ).hexdigest()[:5]

    template = Template(name=template_name, template_type=template_type)
    constructed_template_messages = [
        TemplateMessage(
            role=message.get("role", "assistant"), message=message.get("message", message.get("content")), order_index=i
        )
        for i, message in enumerate(template_messages)
    ]
    version = TemplateVersion(
        template=template,
        template_messages=constructed_template_messages,
        tag=template_tag,
    )

    return version


def register_template(
    template_messages: list[dict[str, Union[str, dict[str, Any]]]],
    template_name: str = None,
    template_tag: str = None,
    template_type=Template.TEMPLATE_TYPE_FORMATTED_STRING,
) -> Template:
    from baserun import Baserun

    if not Baserun.templates:
        Baserun.templates = {}

    if template := Baserun.templates.get(template_name):
        return template

    version = construct_template_version(
        template_messages=template_messages,
        template_name=template_name,
        template_tag=template_tag,
        template_type=template_type,
    )

    request = SubmitTemplateVersionRequest(template_version=version, environment=Baserun.environment)
    response: SubmitTemplateVersionResponse = get_or_create_submission_service().SubmitTemplateVersion(request)

    template = response.template
    if template.name not in Baserun.templates:
        Baserun.templates[template.name] = response.template

    return response.template


async def aregister_template(
    template_messages: list[dict[str, Union[str, dict[str, Any]]]],
    template_name: str = None,
    template_tag: str = None,
    template_type=Template.TEMPLATE_TYPE_FORMATTED_STRING,
) -> Template:
    from baserun import Baserun

    if not Baserun.templates:
        Baserun.templates = {}

    if template := Baserun.templates.get(template_name):
        return template

    version = construct_template_version(
        template_messages=template_messages,
        template_name=template_name,
        template_tag=template_tag,
        template_type=template_type,
    )

    request = SubmitTemplateVersionRequest(template_version=version, environment=Baserun.environment)
    response: SubmitTemplateVersionResponse = await get_or_create_async_submission_service().SubmitTemplateVersion(
        request
    )

    template = response.template
    if template.name not in Baserun.templates:
        Baserun.templates[template.name] = template

    return template



import json
from numbers import Number
from typing import Any, Union

from baserun import Baserun


def convert_to_score(x):
    return x if isinstance(x, int) else int(x is True)


def check(
    name: str,
    actual: Any,
    expected: Any,
    score: Number,
    metadata: dict[str, Any] = None,
    eval_type: str = None,
):
    Baserun.evals._store_eval_data(
        name=name,
        eval_type=eval_type or "match",
        result=json.dumps(actual),
        score=score,
        submission=json.dumps(expected),
        payload=metadata or {},
    )

    return actual


def check_equals(
    name: str,
    actual: str,
    expected: Union[str, list[str]],
    metadata: dict[str, Any] = None,
):
    expected_list = [expected] if isinstance(expected, str) else expected
    result = any(actual == item for item in expected_list)

    return check(
        name=name,
        eval_type="match",
        metadata=metadata,
        actual=actual,
        expected=expected,
        score=1 if result else 0,
    )


def check_includes(
    name: str,
    actual: str,
    expected: Union[str, list[str]],
    metadata: dict[str, Any] = None,
):
    expected_list = [expected] if isinstance(expected, str) else expected
    result = any(actual in item for item in expected_list)

    return check(
        name=name,
        eval_type="match",
        metadata=metadata,
        actual=actual,
        expected=expected,
        score=1 if result else 0,
    )



import logging

from baserun.grpc import (
    get_or_create_submission_service,
    get_or_create_async_submission_service,
)
from baserun.v1.baserun_pb2 import SubmitUserRequest, EndUser

logger = logging.getLogger(__name__)


def submit_user(identifier: str):
    user = EndUser(identifier=identifier)
    user_request = SubmitUserRequest(user=user)
    try:
        get_or_create_submission_service().SubmitUser.future(user_request)
    except Exception as e:
        if hasattr(e, "details"):
            logger.warning(f"Failed to submit user to Baserun: {e.details()}")
        else:
            logger.warning(f"Failed to submit user to Baserun: {e}")


async def asubmit_user(identifier: str):
    user = EndUser(identifier=identifier)
    user_request = SubmitUserRequest(user=user)
    try:
        await get_or_create_async_submission_service().SubmitUser(user_request)
    except Exception as e:
        if hasattr(e, "details"):
            logger.warning(f"Failed to submit user to Baserun: {e.details()}")
        else:
            logger.warning(f"Failed to submit user to Baserun: {e}")



import collections.abc
import json
import logging
from abc import abstractmethod
from datetime import datetime
from random import randint
from typing import Callable, Union
from uuid import UUID

import httpx
from openai._base_client import BaseClient
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion import Choice
from opentelemetry.instrumentation.instrumentor import (
    BaseInstrumentor as OTelInstrumentor,
)
from opentelemetry.sdk.trace import _Span

from baserun.v1.baserun_pb2 import Span, Message, Status, SubmitSpanRequest, ToolCall, ToolFunction

logger = logging.getLogger(__name__)


def spy_on_process_response_data(original_method):
    from baserun import Baserun

    def wrapper(self, *args, **kwargs):
        result: ChatCompletionChunk = original_method(self, *args, **kwargs)
        if not isinstance(result, ChatCompletionChunk):
            return result

        try:
            response = kwargs.get("response")

            if not hasattr(response, "deltas"):
                setattr(response, "deltas", [])

            response.deltas.append(result)
            first_delta = result.choices[0].delta

            # Stream has ended, compile the deltas and submit
            if first_delta.content is None and first_delta.function_call is None and first_delta.tool_calls is None:
                content = ""
                function_name = ""
                function_args = ""
                tool_calls = []
                for delta in response.deltas:
                    if new_content := delta.choices[0].delta.content:
                        content += new_content

                    if new_function := delta.choices[0].delta.function_call:
                        if new_function.name:
                            function_name += new_function.name
                        if new_function.arguments:
                            function_args += new_function.arguments

                    if new_tool_calls := delta.choices[0].delta.tool_calls:
                        for new_tool_call in new_tool_calls:
                            while len(tool_calls) <= new_tool_call.index:
                                tool_calls.append(ToolCall(function=ToolFunction()))

                            tool_call = tool_calls[new_tool_call.index]

                            if new_tool_call.id:
                                tool_call.id = new_tool_call.id

                            if new_tool_call.type:
                                tool_call.type += new_tool_call.type

                            if new_tool_call.function.name:
                                tool_call.function.name += new_tool_call.function.name

                            if new_tool_call.function.arguments:
                                tool_call.function.arguments += new_tool_call.function.arguments

                first_delta = response.deltas[0]

                completion_message = Message(
                    # Role is set on the first delta
                    role=first_delta.choices[0].delta.role,
                    # Content is the aggregate of all deltas
                    content=content,
                    # Finish reason is on the last delta
                    finish_reason=result.choices[0].finish_reason,
                )
                if function_name:
                    completion_message.function_call = json.dumps({"name": function_name, "arguments": function_args})

                if tool_calls:
                    for tool_call in tool_calls:
                        completion_message.tool_calls.append(tool_call)

                span_request = response._span_request
                span = span_request.span
                span.completions.append(completion_message)
                span.completion_id = result.id
                span.model = result.model
                # 1 token per delta, minus 2 (first = empty content, last = None content)
                span.completion_tokens = len(response.deltas) - 2
                # TODO: How does one count prompt tokens here?
                span.prompt_tokens = 0
                span.total_tokens = span.completion_tokens + span.prompt_tokens

                # TODO: Templates

                Baserun.exporter_queue.put(span_request)
        except BaseException as e:
            logger.warning(f"Failed to collect span for Baserun: {e}")
            pass

        return result

    return wrapper


def compile_tool_calls(choice: Choice) -> list[ToolCall]:
    calls = []
    if not choice.message.tool_calls:
        return calls

    for call in choice.message.tool_calls:
        calls.append(
            ToolCall(
                id=call.id,
                type=call.type,
                function=ToolFunction(name=call.function.name, arguments=call.function.arguments),
            )
        )

    return calls


def find_template_match(messages: list[Message]) -> Union[str, None]:
    from baserun import Baserun

    if not messages:
        return None

    message_contents = [message.content for message in messages]

    for template_name, formatted_templates in Baserun.formatted_templates.items():
        for formatted_template in formatted_templates:
            # FIXME? What if there are multiple matches? Maybe check for the highest # of messages
            if all(message in message_contents for message in formatted_template):
                return template_name

    return None


def spy_on_process_response(original_method):
    def wrapper(self, *args, **kwargs):
        from baserun import Baserun, get_template
        import openai

        start_time = datetime.utcnow()
        result: ChatCompletion = original_method(self, *args, **kwargs)
        end_time = datetime.utcnow()
        response: httpx.Response = kwargs.get("response")
        if response:
            try:
                request: httpx.Request = response.request
                parsed_request = json.loads(request.content)
                current_run = Baserun.current_run()
                if not current_run:
                    # TODO handle this case
                    return result

                prompt_messages = [
                    Message(
                        role=message.get("role"),
                        content=message.get("content"),
                    )
                    for message in parsed_request.get("messages", [])
                ]

                if hasattr(result, "choices"):
                    completion_messages = [
                        Message(
                            role=choice.message.role,
                            content=choice.message.content,
                            finish_reason=choice.finish_reason,
                            function_call=choice.message.function_call,
                            tool_calls=compile_tool_calls(choice),
                            system_fingerprint=result.system_fingerprint,
                        )
                        for choice in result.choices
                    ]
                else:
                    # Streaming response, will set the completion data later
                    completion_messages = []

                x_request_id = response.headers.get("x-request-id")
                span = Span(
                    # TODO: Message for non-200
                    status=Status(message="ok", code=response.status_code),
                    prompt_messages=prompt_messages,
                    completions=completion_messages,
                )

                matched_template = find_template_match(prompt_messages)
                if matched_template and (template := get_template(matched_template)):
                    span.template_id = template.id

                if tools := parsed_request.get("tools"):
                    span.tools = json.dumps(tools)

                if functions := parsed_request.get("functions"):
                    span.functions = json.dumps(functions)

                if x_request_id:
                    span.span_id = int.from_bytes(bytes.fromhex(x_request_id[:8]), "big")
                    span.x_request_id = x_request_id
                else:
                    span.span_id = randint(0, 9999999999999999)

                span.run_id = current_run.run_id
                span.trace_id = UUID(current_run.run_id).bytes
                span.name = "openai.chat"
                span.vendor = "openai"
                span.request_type = "chat"
                span.start_time.FromDatetime(start_time)
                span.end_time.FromDatetime(end_time)

                span.api_type = openai.api_type or "open_ai"
                span.api_base = openai.base_url or "https://api.openai.com/v1"
                span.stream = parsed_request.get("stream", False)

                if max_tokens := parsed_request.get("max_tokens"):
                    span.max_tokens = max_tokens
                if temperature := parsed_request.get("temperature"):
                    span.temperature = temperature
                if top_p := parsed_request.get("top_p"):
                    span.top_p = top_p
                if top_k := parsed_request.get("top_k"):
                    span.top_k = top_k
                if frequency_penalty := parsed_request.get("frequency_penalty"):
                    span.frequency_penalty = frequency_penalty
                if presence_penalty := parsed_request.get("presence_penalty"):
                    span.presence_penalty = presence_penalty
                if n := parsed_request.get("n"):
                    span.n = n
                if logit_bias := parsed_request.get("logit_bias"):
                    span.logit_bias = logit_bias
                if logprobs := parsed_request.get("logprobs"):
                    span.logprobs = logprobs
                if stop := kwargs.get("stop"):
                    if isinstance(stop, str):
                        span.stop = [stop]
                    else:
                        span.stop = stop
                if echo := parsed_request.get("echo"):
                    span.echo = echo
                if suffix := parsed_request.get("suffix"):
                    span.suffix = suffix
                if best_of := parsed_request.get("best_of"):
                    span.best_of = best_of
                if user := parsed_request.get("user"):
                    span.user = user
                if function_call := parsed_request.get("function_call"):
                    span.function_call = json.dumps(function_call)
                if tool_choice := parsed_request.get("tool_choice"):
                    span.tool_choice = json.dumps(tool_choice)

                # Non-streaming
                if hasattr(result, "model"):
                    span.model = result.model
                    span.total_tokens = result.usage.total_tokens
                    span.completion_tokens = result.usage.completion_tokens
                    span.prompt_tokens = result.usage.prompt_tokens
                    span.completion_id = result.id

                    span_request = SubmitSpanRequest(span=span, run=current_run)
                    # TODO: Templates
                    Baserun.exporter_queue.put(span_request)
                    return result
                # Streaming- will get submitted in `spy_on_process_response_data` after streaming finishes
                else:
                    span_request = SubmitSpanRequest(span=span, run=current_run)
                    setattr(response, "_span_request", span_request)

            except BaseException as e:
                logger.warning(f"Failed to collect span for Baserun: {e}")
                pass

        return result

    return wrapper


original_methods = {}


def instrument():
    global original_methods
    if original_methods:
        return

    original_methods = {
        "_process_response": BaseClient._process_response,
        "_process_response_data": BaseClient._process_response_data,
    }
    BaseClient._process_response = spy_on_process_response(BaseClient._process_response)
    BaseClient._process_response_data = spy_on_process_response_data(BaseClient._process_response_data)


def uninstrument():
    BaseClient._process_response = original_methods["_process_response"]
    BaseClient._process_response_data = original_methods["_process_response_data"]


class BaseInstrumentor(OTelInstrumentor):
    original_methods: dict[str, Callable] = None

    @staticmethod
    @abstractmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        pass

    @staticmethod
    @abstractmethod
    def async_generator_wrapper(original_generator: collections.abc.AsyncIterator, span: _Span):
        pass

    def _instrument(self, **kwargs):
        self.original_methods = {
            "_process_response": BaseClient._process_response,
            "_process_response_data": BaseClient._process_response_data,
        }
        BaseClient._process_response = spy_on_process_response(BaseClient._process_response)
        BaseClient._process_response_data = spy_on_process_response_data(BaseClient._process_response_data)

    def _uninstrument(self, **kwargs):
        BaseClient._process_response = self.original_methods["_process_response"]
        BaseClient._process_response_data = self.original_methods["_process_response_data"]



import collections.abc
import logging
from typing import TYPE_CHECKING, Collection

from opentelemetry.sdk.trace import _Span

from baserun.instrumentation.base_instrumentor import BaseInstrumentor

logger = logging.getLogger(__name__)

_instruments = ("openai >= 0.27.0",)
__version__ = "0.1.0"

if TYPE_CHECKING:
    pass


class OpenAIInstrumentor(BaseInstrumentor):
    """An instrumentor for OpenAI's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    @staticmethod
    def generator_wrapper(original_generator: collections.abc.Iterator, span: _Span):
        for value in original_generator:
            OpenAIInstrumentor._handle_generator_value(value, span)

            yield value

    @staticmethod
    async def async_generator_wrapper(original_generator: collections.abc.AsyncIterator, span: _Span):
        logged_failure = False
        async for value in original_generator:
            try:
                OpenAIInstrumentor._handle_generator_value(value, span)
            except Exception as e:
                # Avoid spamming failures for each token
                if not logged_failure:
                    logger.info(f"Baserun couldn't handle generator value {value}: {e}")
                    logged_failure = True

            yield value

    @staticmethod
    def _handle_generator_value(value: "OpenAIObject", span: _Span):
        # Currently we only support one choice in streaming responses
        choice = value.choices[0]

        if hasattr(choice, "delta"):
            # Chat
            delta = choice.delta
            role = delta.role
            new_content = delta.content
            new_function_call: "OpenAIObject" = delta.function_call
        else:
            # Completion
            role = None
            new_content = choice.text
            new_function_call = None

        prefix = f"0"
        content_attribute = f"{prefix}.content"
        function_name_attribute = f"{prefix}.function_name"
        function_arguments_attribute = f"{prefix}.function_arguments"

        if role:
            span.set_attribute(f"{prefix}.role", role)

        if new_content:
            content = span.attributes.get(content_attribute, "")
            span.set_attribute(content_attribute, content + new_content)

        if new_function_call:
            function_name = span.attributes.get(function_name_attribute, "")
            if name_delta := new_function_call.name:
                span.set_attribute(function_name_attribute, function_name + name_delta)

            function_arguments = span.attributes.get(function_arguments_attribute, "")
            if arguments_delta := new_function_call.arguments:
                span.set_attribute(function_arguments_attribute, function_arguments + arguments_delta)

        if ((new_content is None and not new_function_call) or choice.finish_reason) and span.is_recording():
            span.end()
